/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2009 Novell, Inc.
 */

#include <string.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <nm-device-ethernet.h>
#include <nm-device-wifi.h>
#include <nm-gsm-device.h>
#include <nm-cdma-device.h>

#include "nmn-status-icon.h"
#include "nmn-icon-cache.h"

#define ACTIVATION_STEPS 6

G_DEFINE_TYPE (NmnStatusIcon, nmn_status_icon, GTK_TYPE_STATUS_ICON)

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NMN_TYPE_STATUS_ICON, NmnStatusIconPrivate))

typedef struct {
    NmnStatusIcon *icon;
    void (*destroy) (gpointer);
} NmnIconHandler;

typedef struct {
    NMClient *client;
    char *current_image;
    gboolean active;

    GSList *ac_list;
    NmnIconHandler *ac_icon_handler;
} NmnStatusIconPrivate;

GtkStatusIcon *
nmn_status_icon_new (void)
{
    return GTK_STATUS_ICON (g_object_new (NMN_TYPE_STATUS_ICON, NULL));
}

static void
update_icon (NmnStatusIcon *self, const char *image)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);
    char *real_image;
    GdkPixbuf *pixbuf;

    real_image = g_strconcat (image, priv->active ? "-active" : "-normal", NULL);
    pixbuf = nmn_icon_cache_get (real_image);
    g_free (real_image);

    if (pixbuf) {
        if (priv->current_image == NULL || strcmp (priv->current_image, image)) {
            g_free (priv->current_image);
            priv->current_image = g_strdup (image);
        }

        gtk_status_icon_set_from_pixbuf (GTK_STATUS_ICON (self), pixbuf);
    }
}

void
nmn_status_icon_set_active (NmnStatusIcon *self,
                            gboolean active)
{
    NmnStatusIconPrivate *priv;

    g_return_if_fail (NMN_IS_STATUS_ICON (self));

    priv = GET_PRIVATE (self);
    if (priv->active != active) {
        priv->active = active;
        update_icon (self, priv->current_image);
    }
}

static void
icon_handler_destroy (NmnIconHandler *handler)
{
    g_return_if_fail (handler != NULL);

    if (handler->destroy)
        handler->destroy (handler);
    else
        g_warning ("Missing destroy method");
}

/* Pending activation animation handler */

typedef struct {
    NmnIconHandler parent;
    guint step;
    guint animation_id;
    guint animation_index;
} NmnPendingActivationHandler;

static void
icon_handler_pending_activation_destroy (gpointer data)
{
    NmnPendingActivationHandler *handler = (NmnPendingActivationHandler *) data;

    g_source_remove (handler->animation_id);
    g_free (handler);
}

static gboolean
activation_animation (gpointer data)
{
    NmnPendingActivationHandler *handler = (NmnPendingActivationHandler *) data;
    char *image;

    if (++handler->step > ACTIVATION_STEPS)
        handler->step = 1;

    image = g_strdup_printf ("nm-progress-working-%02d", handler->step);
    update_icon (handler->parent.icon, image);
    g_free (image);

    return TRUE;
}

static NmnIconHandler *
icon_handler_pending_activation_new (NmnStatusIcon *icon)
{
    NmnPendingActivationHandler *handler;

    handler = g_new0 (NmnPendingActivationHandler, 1);
    handler->parent.icon = icon;
    handler->parent.destroy = icon_handler_pending_activation_destroy;

    handler->animation_id = g_timeout_add (200, activation_animation, handler);
    activation_animation (handler);

    return (NmnIconHandler *) handler;
}


/* Ethernet handler */

static NmnIconHandler *
icon_handler_ethernet_new (NmnStatusIcon *icon)
{
    NmnIconHandler *handler;

    handler = g_new0 (NmnIconHandler, 1);
    handler->destroy = g_free;
    update_icon (icon, "nm-device-wired");

    return handler;
}


/* Wifi handler */

typedef struct {
    NmnIconHandler parent;

    NMDevice *device;
    gulong active_ap_id;

    NMAccessPoint *ap;
    gulong ap_strength_id;
} NmnWifiHandler;

static void
icon_handler_wifi_destroy (gpointer data)
{
    NmnWifiHandler *handler = (NmnWifiHandler *) data;

    g_signal_handler_disconnect (handler->device, handler->active_ap_id);

    if (handler->ap_strength_id)
        g_signal_handler_disconnect (handler->ap, handler->ap_strength_id);

    g_free (handler);
}

static void
ap_strength_changed (NMAccessPoint *ap, GParamSpec *pspec, gpointer user_data)
{
    NmnIconHandler *handler = (NmnIconHandler *) user_data;
    const char *image;
    guint32 strength;

    /* Hack: libnm-glib emits this signal sometimes with AP == NULL.
       Should fix that instead */
    strength = ap ? nm_access_point_get_strength (ap) : 0;
    strength = CLAMP (strength, 0, 100);

    if (strength > 80)
        image = "nm-signal-100";
    else if (strength > 55)
        image = "nm-signal-75";
    else if (strength > 30)
        image = "nm-signal-50";
    else if (strength > 5)
        image = "nm-signal-25";
    else
        image = "nm-signal-00";

    update_icon (handler->icon, image);
}

static void
active_ap_changed (NMDeviceWifi *device, GParamSpec *pspec, gpointer user_data)
{
    NmnWifiHandler *handler = (NmnWifiHandler *) user_data;

    if (handler->ap_strength_id) {
        g_signal_handler_disconnect (handler->ap, handler->ap_strength_id);

        handler->ap = NULL;
        handler->ap_strength_id = 0;
    }

    handler->ap = nm_device_wifi_get_active_access_point (device);
    if (handler->ap) {
        handler->ap_strength_id = g_signal_connect (handler->ap, "notify::" NM_ACCESS_POINT_STRENGTH,
                                                    G_CALLBACK (ap_strength_changed), handler);
    }

    ap_strength_changed (handler->ap, NULL, handler);
}

static NmnIconHandler *
icon_handler_wifi_new (NmnStatusIcon *icon, NMActiveConnection *ac)
{
    const GPtrArray *active_devices;
    NMDevice *device;
    NmnWifiHandler *handler;

    active_devices = nm_active_connection_get_devices (ac);
    if (!active_devices)
        return NULL;

    device = NM_DEVICE (g_ptr_array_index (active_devices, 0));

    handler = g_new0 (NmnWifiHandler, 1);
    handler->parent.icon = icon;
    handler->parent.destroy = icon_handler_wifi_destroy;

    handler->device = device;
    handler->active_ap_id = g_signal_connect (device, "notify::" NM_DEVICE_WIFI_ACTIVE_ACCESS_POINT,
                                              G_CALLBACK (active_ap_changed),
                                              handler);

    active_ap_changed (NM_DEVICE_WIFI (device), NULL, handler);

    return (NmnIconHandler *) handler;
}


/* 3G handler */

static NmnIconHandler *
icon_handler_3g_new (NmnStatusIcon *icon)
{
    NmnIconHandler *handler;

    handler = g_new0 (NmnIconHandler, 1);
    handler->destroy = g_free;
    update_icon (icon, "nm-device-wwan");

    return handler;
}


/* Handler creator */

static NmnIconHandler *
icon_handler_create (NmnStatusIcon *self, NMActiveConnection *ac)
{
    NmnIconHandler *handler = NULL;
    NMActiveConnectionState state;

    state = nm_active_connection_get_state (ac);

    if (state == NM_ACTIVE_CONNECTION_STATE_ACTIVATING)
        handler = icon_handler_pending_activation_new (self);
    else if (state == NM_ACTIVE_CONNECTION_STATE_ACTIVATED) {
        const GPtrArray *active_devices;
        NMDevice *device = NULL;

        active_devices = nm_active_connection_get_devices (ac);
        if (active_devices)
            device = NM_DEVICE (g_ptr_array_index (active_devices, 0));

        if (NM_IS_DEVICE_ETHERNET (device))
            handler = icon_handler_ethernet_new (self);
        else if (NM_IS_DEVICE_WIFI (device))
            handler = icon_handler_wifi_new (self, ac);
        else if (NM_IS_GSM_DEVICE (device) || NM_IS_CDMA_DEVICE (device))
            handler = icon_handler_3g_new (self);
        else
            g_warning ("Unhandled device type: '%s'", G_OBJECT_TYPE_NAME (device));
    }

    return handler;
}

typedef struct {
    NMActiveConnection *ac;
    gulong signal_id;
} AcListItem;

static void
update_best_ac (NmnStatusIcon *self)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);
    GSList *iter;
    NMActiveConnection *best_ac = NULL;

    for (iter = priv->ac_list; iter; iter = iter->next) {
        AcListItem *item = (AcListItem *) iter->data;

        if (nm_active_connection_get_state (item->ac) == NM_ACTIVE_CONNECTION_STATE_UNKNOWN)
            continue;

        if (!best_ac ||
            nm_active_connection_get_state (item->ac) > nm_active_connection_get_state (best_ac) ||
            nm_active_connection_get_default (item->ac) == TRUE)
            best_ac = item->ac;
    }

    if (priv->ac_icon_handler) {
        icon_handler_destroy (priv->ac_icon_handler);
        priv->ac_icon_handler = NULL;
    }

    if (best_ac)
        priv->ac_icon_handler = icon_handler_create (self, best_ac);

    if (!priv->ac_icon_handler)
        update_icon (self, "nm-no-connection");
}

static void
ac_state_changed (NMActiveConnection *ac,
                  GParamSpec *pspec,
                  gpointer user_data)
{
    update_best_ac (NMN_STATUS_ICON (user_data));
}

static void
ac_list_destroy (NmnStatusIcon *self)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);

    while (priv->ac_list) {
        AcListItem *item = (AcListItem *) priv->ac_list->data;

        g_signal_handler_disconnect (item->ac, item->signal_id);
        g_object_unref (item->ac);
        g_free (item);
        priv->ac_list = g_slist_delete_link (priv->ac_list, priv->ac_list);
    }
}

static void
active_connections_changed (NMClient *client,
                            GParamSpec *pspec,
                            gpointer user_data)
{
    NmnStatusIcon *self = NMN_STATUS_ICON (user_data);
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);
    const GPtrArray *ac_list;
    int i;

    /* First, disconnect all previous handlers */
    ac_list_destroy (self);

    /* add signal handlers for each AC */
    if (nm_client_get_manager_running (client))
        ac_list = nm_client_get_active_connections (client);
    else
        ac_list = NULL;

    for (i = 0; ac_list && i < ac_list->len; i++) {
        NMActiveConnection *ac = (NMActiveConnection *) g_ptr_array_index (ac_list, i);
        AcListItem *item;

        item = g_new (AcListItem, 1);
        item->ac = g_object_ref (ac);
        item->signal_id = g_signal_connect (ac, "notify", G_CALLBACK (ac_state_changed), self);
        priv->ac_list = g_slist_prepend (priv->ac_list, item);
    }

    update_best_ac (self);
}

void
nmn_status_icon_set_client (NmnStatusIcon *self,
                            NMClient *client)
{
    NmnStatusIconPrivate *priv;

    g_return_if_fail (NMN_IS_STATUS_ICON (self));
    g_return_if_fail (NM_IS_CLIENT (client));

    priv = GET_PRIVATE (self);
    priv->client = g_object_ref (client);

    g_signal_connect (priv->client, "notify::" NM_CLIENT_ACTIVE_CONNECTIONS,
	                  G_CALLBACK (active_connections_changed),
	                  self);

    active_connections_changed (client, NULL, self);
}

/*****************************************************************************/

static gboolean
set_initial_icon (gpointer data)
{
    update_icon (NMN_STATUS_ICON (data), "nm-no-connection");

    return FALSE;
}

static void
nmn_status_icon_init (NmnStatusIcon *self)
{
    g_idle_add (set_initial_icon, self);
}

static void
finalize (GObject *object)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (object);

    if (priv->ac_icon_handler)
        icon_handler_destroy (priv->ac_icon_handler);

    ac_list_destroy (NMN_STATUS_ICON (object));

    if (priv->client)
        g_object_unref (priv->client);

    G_OBJECT_CLASS (nmn_status_icon_parent_class)->finalize (object);
}

static void
nmn_status_icon_class_init (NmnStatusIconClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);

    g_type_class_add_private (object_class, sizeof (NmnStatusIconPrivate));

    object_class->finalize = finalize;
}
